// Using Statements
using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;

namespace Ziggyware
{
    public class XBoxVideoFeed : DrawableGameComponent
    {

        NetworkSession networkSession;

        PacketWriter packetWriter = new PacketWriter();
        PacketReader packetReader = new PacketReader();        

        int screenWidth;
        int screenHeight;


        
        Color[] colorData;
        int textureWidth;
        int textureHeight;

#if XBOX360

        ResolveTexture2D screenshot;
        int colorDataSent = -1;
        int packetSendSize = 450;

        float throttle = 1.0f;
        float throttleTime = 0.0f;

        RenderTarget2D renderTarget;
        DepthStencilBuffer depthBuffer;
#else
        public Texture2D texture = null;
        int colorDataReceived = -1;
#endif
        int colorDataSize = -2;

        


        SpriteBatch spriteBatch;

        public XBoxVideoFeed(
            Game game,
            int screenWidth,
            int screenHeight,
            int textureWidth,
            int textureHeight,
            NetworkSession session)
            : base(game)
        {
            this.screenWidth = screenWidth;
            this.screenHeight = screenHeight;

            this.textureWidth = textureWidth;
            this.textureHeight = textureHeight;

            colorData = new Color[textureWidth * textureHeight];

            if (session != null)
            {
                this.networkSession = session;
            }
        }


        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
#if XBOX360
            renderTarget = new RenderTarget2D(GraphicsDevice,
                textureWidth, textureHeight, 
                    1, SurfaceFormat.Color);

            depthBuffer = new DepthStencilBuffer(GraphicsDevice,
                    textureWidth, textureHeight, DepthFormat.Depth24Stencil8);

            screenshot =
                new ResolveTexture2D(GraphicsDevice,
                    screenWidth, screenHeight, 
                    1, SurfaceFormat.Color);
#else
            texture = new Texture2D(GraphicsDevice, textureWidth, textureHeight);
#endif

            base.LoadContent();
        }

        protected override void UnloadContent()
        {
#if XBOX360
            renderTarget.Dispose();
            renderTarget = null;

            depthBuffer.Dispose();
            depthBuffer = null;

            screenshot.Dispose();
            screenshot = null;

#endif

            base.UnloadContent();
        }



        // Update

        public override void Update(GameTime gameTime)
        {
            if (networkSession == null)
            {
#if XBOX360
                CreateSession();
#else
                if (Gamer.SignedInGamers.Count == 0)
                {
                    if(Guide.IsVisible == false)
                        Guide.ShowSignIn(4, false);
                }
                else
                {
                    JoinSession();
                }
#endif
            }
            else
            {
                for(int x=0;x<10;x++)
                UpdateNetworkSession(gameTime);
            }


            base.Update(gameTime);
        }



        void CreateSession()
        {
            networkSession = NetworkSession.Create(NetworkSessionType.SystemLink,
                                                   4, 16);
            HookSessionEvents();
        }


        void JoinSession()
        {
            // Search for sessions.
            using (AvailableNetworkSessionCollection availableSessions =
                        NetworkSession.Find(NetworkSessionType.SystemLink,
                                            4, null))
            {
                if (availableSessions.Count == 0)
                {
                    return;
                }

                // Join the first session we found.
                networkSession = NetworkSession.Join(availableSessions[0]);

                HookSessionEvents();
            }

        }


        void HookSessionEvents()
        {
            networkSession.GamerJoined += GamerJoinedEventHandler;
            networkSession.SessionEnded += SessionEndedEventHandler;
            networkSession.GamerLeft += 
                new EventHandler<GamerLeftEventArgs>(networkSession_GamerLeft);
            
        }

        void networkSession_GamerLeft(object sender, GamerLeftEventArgs e)
        {
            colorDataSize = -2;
#if XBOX360
            this.colorDataSent = -1;
#else
            this.colorDataReceived = -1;
#endif
        }


        void GamerJoinedEventHandler(object sender, GamerJoinedEventArgs e)
        {
            colorDataSize = -2;
#if XBOX360
            this.colorDataSent = -1;
#else
            this.colorDataReceived = -1;
#endif
        }

        void SessionEndedEventHandler(object sender, NetworkSessionEndedEventArgs e)
        {
            networkSession.Dispose();
            networkSession = null;
        }

        

        void UpdateNetworkSession(GameTime gameTime)
        {
#if XBOX360

            
            if (networkSession.RemoteGamers.Count <= 0)
            {
                colorDataSize = -2;
                this.colorDataSent = -1;
            }
            else
            {
                if(colorDataSize != -2)
                foreach (LocalNetworkGamer gamer in networkSession.LocalGamers)
                {
                    throttleTime -= (float)gameTime.ElapsedGameTime.TotalSeconds;
                    if (throttleTime <= 0)
                    {
                        throttleTime = throttle;

                        if (this.colorDataSent < 0 ||
                            this.colorDataSize < 0)
                        {
                            this.colorDataSize = this.textureWidth * this.textureHeight;

                            packetWriter.Write(this.textureWidth);
                            packetWriter.Write(this.textureHeight);
                            packetWriter.Write(this.colorDataSize);

                            colorDataSent = 0;

                            gamer.SendData(packetWriter, SendDataOptions.ReliableInOrder);
                        }
                        else if(this.colorDataSent >= 0)
                        {
                            

                            int sendSize = Math.Min(this.colorDataSize - this.colorDataSent,
                                                this.packetSendSize);

                            packetWriter.Write(colorDataSent);
                            packetWriter.Write(sendSize);

                            for (int x = colorDataSent; x < colorDataSent + sendSize; x++)
                            {
                                packetWriter.Write(colorData[x]);
                            }

                            gamer.SendData(packetWriter, SendDataOptions.ReliableInOrder);

                            colorDataSent += sendSize;

                            if (colorDataSent == colorDataSize)
                            {
                                colorDataSize = -2;
                                this.colorDataSent = -1;
                            }
                        }
                    }
                }
            }
#endif

            // Make sure the session has not ended.
            if (networkSession == null)
                return;


            // Pump the underlying session object.
            try
            {
                networkSession.Update();
            }
            catch (Microsoft.Xna.Framework.Net.NetworkException ex)
            {
                NetworkException e = ex;
#if XBOX360
                if (true)
                {
                    packetSendSize -= packetSendSize / 2;
               }
#endif
            }

            if (networkSession == null)
                return;

#if XBOX360
            // Read any packets telling us the positions of remotely controlled tanks.
            foreach (LocalNetworkGamer gamer in networkSession.LocalGamers)
            {
                
                while (gamer.IsDataAvailable)
                {
                    NetworkGamer sender;
                    // Read a single packet from the network.
                    gamer.ReceiveData(packetReader, out sender);
                }
            }
#else
            // Read any packets telling us the positions of remotely controlled tanks.
            foreach (LocalNetworkGamer gamer in networkSession.LocalGamers)
            {


                while (gamer.IsDataAvailable)
                {
                    NetworkGamer sender;

                    // Read a single packet from the network.
                    gamer.ReceiveData(packetReader, out sender);
                    
                    if (packetReader.Length > 0)
                    {
                        if (this.colorDataReceived == -1)
                        {
                            this.textureWidth = packetReader.ReadInt32();
                            this.textureHeight = packetReader.ReadInt32();
                            this.colorDataSize = packetReader.ReadInt32();
                            
                            colorDataReceived = 0;
                        }
                        else
                        {

                            int icolorDataSent = packetReader.ReadInt32();
                            int isendSize = packetReader.ReadInt32();

                            for (int x = 0;
                                     x < isendSize;
                                     x++)
                            {
                                colorData[icolorDataSent + x] = packetReader.ReadColor();
                            }

                            colorDataReceived += isendSize;

                            GraphicsDevice.Textures[0] = null;
    
                            texture.SetData<Color>(colorData);

                            if (colorDataReceived >= colorDataSize)
                            {
                            
                                colorDataReceived = -1;
                                colorDataSize = -2;
                            }
                        }
                    }

                }
            }
#endif
        }



#if XBOX360
        int ibc = 0;
        void GrabBackBuffer()
        {
            if (networkSession == null ||
                networkSession.RemoteGamers.Count <= 0)
                return;

            System.Diagnostics.Debug.WriteLine("Grabbing BBuffer " + (ibc++).ToString());
            GraphicsDevice device = GraphicsDevice;
            {
                device.ResolveBackBuffer(screenshot);
                
                
                device.SetRenderTarget(0, renderTarget);

                DepthStencilBuffer oldDepth = device.DepthStencilBuffer;
                device.DepthStencilBuffer = depthBuffer;
                Viewport vp = device.Viewport;
                vp.Width = textureWidth;
                vp.Height = textureHeight;
                device.Viewport = vp;

                device.Clear(Color.Black);

                spriteBatch.Begin();
                
                spriteBatch.Draw(screenshot, 
                    new Rectangle(0, 0, textureWidth, textureHeight), 
                    Color.White);

                spriteBatch.End();

                device.SetRenderTarget(0, null);
                
                device.DepthStencilBuffer = oldDepth;

                vp.Width = screenWidth;
                vp.Height = screenHeight;
                device.Viewport = vp;

               
                renderTarget.GetTexture().GetData<Color>(this.colorData);

                colorDataSize = colorData.Length;
                colorDataSent = -1;
                
            }
        }
#endif

        // Draw
        public override void Draw(GameTime gameTime)
        {

            

#if WINDOWS
            if (networkSession == null)
            {
            }
            else
            {
                if (texture != null)
                {
                    GraphicsDevice.Clear(Color.CornflowerBlue);
                    spriteBatch.Begin();
                    spriteBatch.Draw(texture, 
                        new Rectangle(0,0,screenWidth,screenHeight), 
                        Color.White);
                    spriteBatch.End();
                }
            }
#else
            if (colorDataSize == -2)
            {
                GrabBackBuffer();
            }
#endif

            base.Draw(gameTime);

            
        }

    }
}
